/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         KqmlRegistration.java
**
** Description:  Contains the KqmlRegistration class that is internal to
**               the KRIL.
**
*/
package kqml;
import java.lang.Thread;

//
// This class is used to support blocking send's (a Kqml message send
// followed by a reply wait) in application threads.
//

/**
 * A class internal to the KRIL.  Not currently documented.
 *
 * @author Peter Stevens
 * @version $Revision: 1.6 $
 * @see KqmlAgent#addLookupThread()
 * @see KqmlAgent#checkTimeouts()
 * @see KqmlAgent#lookupThread()
 * @see KqmlAgent#processConn()
 * @see KqmlAgent#registerThread()
 */
class KqmlRegistration
{
    private KqmlAgent agent    = null;
    private MsgQueue  msgQueue = null;
    private Thread    thread   = null;
    private long      timeout  = 0;
    private static final long PENDING_MSGS_TIME_INTERVAL = 100000;

    // Constructor
    KqmlRegistration(KqmlAgent agent)
    {
        this.agent = agent;
        thread     = Thread.currentThread();
        msgQueue   = new MsgQueue(thread);
        agent.addLookupThread(this);    // for sendWithReply operations 
    }

  protected void makePending(Message msg)
  {
    //
    // Note this thread is waiting for this message
    //
    agent.pendingMsgPutThread(msg, this);

    // Since this message is now on the pendingMsgs queue, it can be
    // accessed by the checkTimeouts() method of the KqmlAgent object
    // which will check to see if it has timed out.  So, we must reset
    // the timeout value for this thread here.  Since we don't want to
    // set that value for real until getExpectedMsg() is called, set it
    // to something big enough that it will allow for the different 
    // threads to catch up to each other.
    this.timeout = System.currentTimeMillis() + PENDING_MSGS_TIME_INTERVAL;

    Log.log(Log.IO+Log.DEBUG, "KqmlRegistration:makePending\n     msgId `" +
	    msg.getKqmlMsgId() + "' for dispatch thread `" +
	    this.getName() + "'");
  }

  protected Message getExpectedMsg(long timeout)
  {
    //
    // The timeout value is computed here as a time in the future
    // to save having to compute it every time the timeouts are
    // checked in KqmlAgent.checkTimeouts().
    //
    this.timeout = System.currentTimeMillis() + timeout;
    Log.log(Log.DEBUG, "getExpectedMsg: set timeout=" + this.timeout +
	    " for thread `" + this.getName() + "'");
    
    //
    // Block this thread, waiting for the message
    //
    return (msgQueue.getExpectedMsg());
  }
  
  protected void putExpectedMsg(Message msg) { msgQueue.putExpectedMsg(msg); }
  
  protected Thread getThread()   { return (thread); }
  
  protected String getName()     { return (thread.getName()); }
  
  protected long   getTimeout()  { return (timeout); }
}
