/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         KrilWrapper.java
**
** Description:  Contains the KrilWrapper class with the following "public"
**               constructors and methods:
**
**                 Constructor to decode the :kril-wrapper from an incoming
**                 message.  There is accesor methods for getting the
**                 input value Object result and the encoded :kril-wrapper
**                 and :content field results.
**
*/
package kqml;
import java.util.StringTokenizer;
import java.util.Vector;

//
// Outbound message functions in this class are internal to the KRIL.
//
// - It generates a kril-wrapper value and content field used for outbound 
//   messages ...
//
// - OR it parses an inbound message's kril-wrapper value and content fields
//   and generates the appropriate object (matching the description in the
//   :kril-wrapper).
//

/**
 * A class to encode and decode <code>:kril-wrapper</code> for outgoing
 * and incoming KQML message's respectively.  The KrilWrapper encoding
 * is executed the Message constructors (though it isn't re-executed
 * automatically following a Message.setContents method call).  There
 * accessor methods provided to fetch encoded <code>:kril-wrapper</code>
 * and <code>:content</code> outputs for outgoing messages (used in Message
 * constructors) and the input value Object from decoding incoming messages.
 *
 * <b>Note:  incoming message KrilWrapper decoding is not executed by the
 * KRIL automatically, as it is for outgoing messages.</b>
 *
 * @author Peter Stevens
 * @version $Revision: 1.5 $
 * @see Message
 */
public class KrilWrapper
{
    private static final String KRIL_WRAPPER    = ":kril-wrapper";
    private static final String LONG_TYPE       = "L";  // for C long
    private static final int    MIN_WRAPPER_LEN = 5;
    private static final String ONE_COUNT       = ":1:";
    private static final String SEPARATOR       = ":";
    private static final String STRING_TYPE     = "S";
    private static final String ULONG_TYPE      = "U";  // for C unsigned long
    private static final long   THIRTY_TWO_BITS = 0xFFFFFFFFL;

    // Used to create an outbound wrapper
    private String      content     = null;     // set for inbound msgs too
    private String      krilWrapper = null;     // set for inbound msgs too

    // Used to parse a inbound wrapper
    private Object      inputVal    = null;

    // Constructors ... the first set take a reply value and generate
    //                  the appropriate outbound kril-wrapper
    protected KrilWrapper(String s)
    {
        content     = s;
        krilWrapper = STRING_TYPE + ONE_COUNT + content.length();
    }

    protected KrilWrapper(int i)
    {
        content     = Integer.toString(i);
        krilWrapper = LONG_TYPE   + SEPARATOR + "1";
    }

    protected KrilWrapper(long l) throws InvalidLongInputException
    {
        if ((l > THIRTY_TWO_BITS) || ((l < 0) && ((-l) > THIRTY_TWO_BITS)))
            throw new InvalidLongInputException("Reply value " + l +
                        " uses too many bits (longer than 32 bits)" +
                        "[" + Long.toString(THIRTY_TWO_BITS) + "]");

        content     = Long.toString(l);
        krilWrapper = ULONG_TYPE  + SEPARATOR + "1";
    }

    protected KrilWrapper(String [] array)
    {
        if (array.length > 0)
        {
            content     = array[0];
            krilWrapper = STRING_TYPE + SEPARATOR + array.length + SEPARATOR +
                                                        content.length();
            for (int i = 1; i < array.length; ++i)
            {
                String s     = array[i];
                content     += " " + s;
                krilWrapper += SEPARATOR + s.length();
            }
            content += " ";     // for compatibility with the "C" kqml_wrapper
        }
    }

    //
    // The vector here is assumed to a homogeneous vector of strings
    // for compatibility with the current KqmlSql class of functions.
    // This may eventually be deprecated by the String array version
    // for stronger type safety.
    //
    protected KrilWrapper(Vector vec)
    {
        if (vec.size() > 0)
        {
            content      = (String)vec.firstElement();
            krilWrapper  = STRING_TYPE + SEPARATOR + vec.size() + SEPARATOR +
                                                        content.length();
            for (int i = 1; i < vec.size(); ++i)
            {
                String s     = (String)vec.elementAt(i);
                content     += " " + s;
                krilWrapper += SEPARATOR + s.length();
            }
            content += " ";     // for compatibility with the "C" kqml_wrapper
        }
    }

    protected KrilWrapper(int [] array)
    {
        if (array.length > 0)
        {
            content      = Long.toString(array[0]);
            krilWrapper  = LONG_TYPE + SEPARATOR + array.length;
            for (int i = 1; i < array.length; ++i)
            {
                String s     = Long.toString(array[i]);
                content     += " " + s;
            }
            content += " ";     // for compatibility with the "C" kqml_wrapper
        }
    }

    protected KrilWrapper(long [] array) throws InvalidLongInputException
    {
        if (array.length > 0)
        {
            long l             = array[0];

            if ((l > THIRTY_TWO_BITS) || ((l < 0) && ((-l) > THIRTY_TWO_BITS)))
                throw new InvalidLongInputException("Reply value " + l +
                            " uses too many bits (longer than 32 bits)" +
                            "[" + Long.toString(THIRTY_TWO_BITS) + "]");

            content      = Long.toString(array[0]);
            krilWrapper  = ULONG_TYPE + SEPARATOR + array.length;
            for (int i = 1; i < array.length; ++i)
            {
                l = array[i];
                if ((l > THIRTY_TWO_BITS) || ((l<0) && ((-l)>THIRTY_TWO_BITS)))
                    throw new InvalidLongInputException("Reply value " + l +
                                " uses too many bits (longer than 32 bits)");

                String s     = Long.toString(l);
                content     += " " + s;
            }
            content += " ";     // for compatibility with the "C" kqml_wrapper
        }
    }

    /**
     * This constructor parses in an in-bound kril-wrapper generating
     * the appropriate result into the "inputVal" Object.  The object
     * will be one of String, String[], Integer, Integer[], Long, or
     * Long[].
     *
     * @param m the incoming Message which needs kril wrapper decoding.
     * @see Message
     * @exception InvalidKrilWrapperException generated if the incoming
     * message's <code>:kril-wrapper</code> is malformed.
     */
    public KrilWrapper(Message m) throws InvalidKrilWrapperException
    {
        int              count       = 0;
        boolean          quoted      = false;
        StringTokenizer  token       = null;
        String           type        = null;

        content     = m.getContents();
        krilWrapper = m.getKrilWrapper();
        if (krilWrapper != null)
        {
            try     { quoted = (krilWrapper.charAt(0) == '"'); }
            // this catch SHOULD never happen
            catch   (StringIndexOutOfBoundsException e)
            { 
                throw new InvalidKrilWrapperException("kril-wrapper " +
                    "`" + krilWrapper + "' is malformed");
            }

            int     startChar   = quoted ? 1 : 0;
            
            try
            {
                //
                // used to trim trailing `"', if string started with a quote 
                //
                int lastChar = krilWrapper.length() - startChar;

                token = new StringTokenizer(
                                krilWrapper.substring(startChar, lastChar),
                                        SEPARATOR);
            }
            // this catch SHOULD never happen
            catch   (StringIndexOutOfBoundsException e)
            { 
                throw new InvalidKrilWrapperException("kril-wrapper " +
                    "`" + krilWrapper + "' is malformed");
            }

            if (token.hasMoreTokens())
                type = token.nextToken();
            else
                throw new InvalidKrilWrapperException("kril-wrapper " + "`" +
                    krilWrapper + "' is missing TYPE field colon separator");

            //
            // The ordering of these "if's" is in expected likelihood of
            // occurence.
            //
            if (type.equalsIgnoreCase(STRING_TYPE))
            {
                inputVal = processStrings(token, content);
            }
            else if (type.equalsIgnoreCase(LONG_TYPE))
            {
                inputVal = processLongs(token, content);
            }
            else if (type.equalsIgnoreCase(ULONG_TYPE))
            {
                inputVal = processUlongs(token, content);
            }
            else
            {
                throw new InvalidKrilWrapperException("kril-wrapper type `" +
                    type + "' isn't supported");
            }
        }
    }

    private Object processLongs(StringTokenizer t, String c) throws
                                            InvalidKrilWrapperException
    {
        int count   = getCount(t);
        Object rval = null;

        if (count == 1)
        {
            try
            {
                rval = new Integer(c.trim());
            }
            catch (NumberFormatException e)
            {
                throw new InvalidKrilWrapperException(
                    "kril-wrapper LONG value" + "`" + c +
                    "' is not a valid integer");
            }
        }
        else
        {
            int         contentOffset = 0;
            int         endOffset     = 0;

            rval = new Integer[count];
            for (int i = 0; i < count; ++i)
            {
                String sub  = null;
                try
                {
                    // Uses ' ' as a delimiter, no lengths needed
                    endOffset = c.indexOf(' ', contentOffset);
                    sub       = c.substring(contentOffset, endOffset);
                    ((Integer [])rval)[i] = new Integer(sub.trim());
                    contentOffset = endOffset + 1;
                }
                catch (NumberFormatException e)
                {
                    throw new InvalidKrilWrapperException(
                        "kril-wrapper content value" + "`" + sub +
                        "' is not a valid integer");
                }
                catch (StringIndexOutOfBoundsException e)
                {
                    throw new InvalidKrilWrapperException(
                        "kril-wrapper content length too short at offset " +
                        contentOffset + " in field " + (i + 1));
                }
            }
        }
        return (rval);
    }

    private Object processStrings(StringTokenizer t, String c) throws
                                            InvalidKrilWrapperException
    {
        int count   = getCount(t);
        Object rval = null;

        if (count != t.countTokens())
            throw new InvalidKrilWrapperException("kril-wrapper count " +
                count + " doesn't match tokenized field count of " +
                t.countTokens());

        if (count == 1)
        {
            rval = c;
        }
        else
        {
            int         contentOffset = 0;
            int         length = 0;

            rval = new String[count];
            for (int i = 0; i < count; ++i)
            {
                if (t.hasMoreTokens())
                {
                    String next = t.nextToken();
                    String sub  = null;
                    try
                    {
                        length = Integer.parseInt(next);
                    }
                    catch (NumberFormatException e)
                    {
                        throw new InvalidKrilWrapperException(
                            "kril-wrapper format length value" + "`" + next +
                            "' is not a valid integer");
                    }
                    try
                    {
                        sub = c.substring(contentOffset, contentOffset+length);
                        ((String [])rval)[i] = sub;
                        contentOffset += length + 1;
                    }
                    catch (StringIndexOutOfBoundsException e)
                    {
                        throw new InvalidKrilWrapperException(
                            "kril-wrapper content length too short at offset " +
                            contentOffset + " in field " + (i + 1));
                    }
                }
                else
                    // Should NEVER happen - don't have kril-wrapper value
                    // here to put in exception string
                    throw new InvalidKrilWrapperException(
                        "kril-wrapper STRING value" + "`" + c +
                        "' has fewer parsed format lengths than specified");
            }
        }
        return (rval);
    }

    // "U", Ulong type is treated as a Java type of "Long"
    private Object processUlongs(StringTokenizer t, String c) throws
                                            InvalidKrilWrapperException
    {
        int count   = getCount(t);
        Object rval = null;

        if (count == 1)
        {
            try
            {
                rval = new Long(c.trim());
            }
            catch (NumberFormatException e)
            {
                throw new InvalidKrilWrapperException(
                    "kril-wrapper ULONG value" + "`" + c +
                    "' is not a valid integer");
            }
        }
        else
        {
            int         contentOffset = 0;
            int         endOffset     = 0;

            rval = new Long[count];
            for (int i = 0; i < count; ++i)
            {
                String sub  = null;
                try
                {
                    // Uses ' ' as a delimiter, no lengths needed
                    endOffset = c.indexOf(' ', contentOffset);
                    sub       = c.substring(contentOffset, endOffset);
                    ((Long [])rval)[i] = new Long(sub.trim());
                    contentOffset = endOffset + 1;
                }
                catch (NumberFormatException e)
                {
                    throw new InvalidKrilWrapperException(
                        "kril-wrapper content value" + "`" + sub +
                        "' is not a valid integer");
                }
                catch (StringIndexOutOfBoundsException e)
                {
                    throw new InvalidKrilWrapperException(
                        "kril-wrapper content length too short at offset " +
                        contentOffset + " in field " + (i + 1));
                }
            }
        }
        return (rval);
    }

    private int getCount(StringTokenizer t) throws
                                            InvalidKrilWrapperException
    {
        int count = 0;

        if (t.hasMoreTokens())
        {
            String countString = t.nextToken();
            try
            {
                count = Integer.parseInt(countString);
            }
            catch (NumberFormatException e)
            { 
                throw new InvalidKrilWrapperException("kril-wrapper count" +
                    "`" + countString + "' is not a valid integer");
            } 
        }
        else
            throw new InvalidKrilWrapperException("kril-wrapper " + "`" +
                krilWrapper + "' is missing COUNT field colon separator");

        return (count);
    }

    /**
     * Accesor to fetch the <code>:content</code> value of an encoded or decoded
     * KrilWrapper.
     *
     * @return the KrilWrapper encoded content field
     */
    public String getContent()     { return (content); };

    /**
     * Accesor to fetch the <code>:kril-wrapper</code> value of an encoded
     * or decoded KrilWrapper.
     *
     * @return the KrilWrapper encoded kril-wrapper field
     */
    public String getKrilWrapper() { return (krilWrapper); };

    /**
     * Accesor to fetch the KrilWrapper decoded Object for incoming message.
     * The object will be one of String, String[], Integer, Integer[], Long,
     * or Long[].
     *
     * @return the KrilWrapper decoded object.
     */
    public Object getInputVal()    { return (inputVal); };
}
