/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         MsgQueue.java
**
** Description:  Contains the MsgQueue class that is internal to the KRIL.
**
*/
package kqml;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Date;

/**
 * A class internal to the KRIL.
 *
 * This class is used to implement message queueing for expected an KQML
 * message.
 *
 * A message queue is associated with a thread which will be suspended
 * for a getExpectedMsg call and resumed if putExpectedMsg is called.
 *
 * @author Peter Stevens
 * @version $Revision: 1.6 $
 */

class MsgQueue
{
    private Vector    expected = null;
    private Thread    thread = null;
    private long      timeout = 0;

    // Constructor
    MsgQueue(Thread thread)
    {
        expected  = new Vector();
        this.thread = thread;
    }

    protected void putExpectedMsg(Message msg)
    {
        // Put a message on the thread's message queue
        Log.log(Log.READ+Log.DEBUG, "putExpectedMsg called");
        synchronized (expected)
        {
            expected.insertElementAt(msg, 0);
        }
        // Resume the thread (no-op if the queue is already running)
        Log.log(Log.THREAD, "RESUME for expected message thread: " +
                thread.getName());
        thread.resume();
    }

    protected Message getExpectedMsg()
    {
      Message rval = readQueue();
      
      //
      // If this thread's message queue expected message isn't present,
      // block.
      // --- the suspend MUST take place outside of the "monitor"
      //
      if (rval == null)
        {
	  Log.log(Log.THREAD,"SUSPEND for expected msg: " + thread.getName());
	  //
	  // Check the queue again JUST before suspending, because the "log"
	  // call takes real-time and we might get the processor taken from
	  // us by the MsgReceive or KqmlAdmin tasks.
	  //
	  if ((rval=readQueue()) == null)
	    thread.suspend();
	  Log.log(Log.THREAD, "RESUMED expected message: "+ thread.getName());
        }
      
      if (rval == null)
	rval = readQueue();
      
      if (rval == null)
	{
	  Log.log(Log.READ+Log.DEBUG, "Timeout for expected message ");
	}
      else
        {
	  Log.log(Log.READ+Log.DEBUG, "Expected message is `" +
		  rval.getPerformative() + "'\n     kqml-msg-id `" +
		  rval.getKqmlMsgId() + "'");
        }
      
      return (rval);
    }
  
    private Message readQueue()
    {
        Message rval;

        synchronized (expected)
        {
            if (expected.isEmpty())
                rval = null;
            else
            {
                // a NoSuchElementException SHOULD NEVER occur here
                rval = (Message)expected.firstElement();
                expected.removeElementAt(0);
            }
        }
        return (rval);
    }

    private Message removeTimedOutMsg()
    {
        Message rval;

        synchronized (expected)
        {
            if (expected.isEmpty())
                rval = null;
            else
            {
                // a NoSuchElementException SHOULD NEVER occur here
                rval = (Message)expected.lastElement();
		// expected.removeElementAt(expected.elementCount - 1);
            }
        }
        return (rval);
    }
}
