/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         Reply.java
**
** Description:  Contains the Reply class with the following "public"
**               constructors and methods:
**
**                 Constructor to generate a reply with a performative
**                 and :kril-wrapper output for the following :content
**                 types:  String, int, int[], long, and long[].
**                 Static methods are provided to return Reply
**                 objects with a canned performative and one of the
**                 supported data types (String, String[], int, int[], etc.).
**                 "reply", "tell", "error", and "sorry" are the
**                 available canned Reply performatives.  The latter
**                 two only have methods for a String content value.
**                 Finally, selector methods are provide for the content, 
**                 and performative.
*/
package kqml;
import java.util.Enumeration;
import java.util.Hashtable;
/**
 * This class is used principally by a KQML message handler in a KQML
 * server agent to return a reply to a sender requesting a reply.
 *
 * It supplies Constructor to generate a reply with a performative
 * and <code>:kril-wrapper</code> output for the following 
 * <code>:content</code> types:  String, int, int[], long, and long[].
 * Static methods are provided to return Reply objects with a canned
 * performative and one of the supported data types (String, String[],
 * int, int[], etc.).  <i>reply</i>, <i>tell</i>, <i>error</i>, and
 * <i>sorry</i> are the available canned Reply performatives.  The
 * latter two only have methods for a String content value.  Finally,
 * selector methods are provide for the content and  performative.
 * values of <i>reply</i> are supplied.
 *
 * @author Peter Stevens
 * @version $Revision: 1.5 $
 * @see Handler
 * @see KrilWrapper
 * @see Message
 */
public class Reply implements KqmlKeywords
{
    private Hashtable   keywordValues = new Hashtable();
    private String      performative  = null;

    /**
     * Returns a Reply instance.
     *
     * @param performative  the reply message's performative 
     * @param content       the reply message's content
     *
     * @see KrilWrapper
     */
    public Reply(String performative, String content)
    {
        this.performative   = performative;
        keywordValues.put(CONTENT,      content);
        keywordValues.put(KRIL_WRAPPER,
                                    new KrilWrapper(content).getKrilWrapper());
    }

    //
    // In these constructors the KrilWrapper object constructor will
    // generate the "content" string and generate the matching
    // ":kril-wrapper" description string
    //

    /**
     * Returns a Reply instance.
     *
     * @param performative  the reply message's performative 
     * @param strings       the reply message's string array to concatenated
     *
     * @see KrilWrapper
     */
    public Reply(String performative, String [] strings)
    {
        this.performative   = performative;
        KrilWrapper wrapper = new KrilWrapper(strings);
        keywordValues.put(CONTENT,      wrapper.getContent());
        keywordValues.put(KRIL_WRAPPER, wrapper.getKrilWrapper());
    }

    /**
     * Returns a Reply instance.
     *
     * @param performative  the reply message's performative 
     * @param i             the reply message's integer value to be converted
     * from an integer to a string with the data type indicated in the
     * <code>:kril-wrapper</code> field of the reply message
     *
     * @see KrilWrapper
     */
    public Reply(String performative, int i)
    {
        this.performative   = performative;
        KrilWrapper wrapper = new KrilWrapper(i);
        keywordValues.put(CONTENT,      wrapper.getContent());
        keywordValues.put(KRIL_WRAPPER, wrapper.getKrilWrapper());
    }

    /**
     * Returns a Reply instance.
     *
     * @param performative  the reply message's performative 
     * @param i             the reply message's integer array value to be
     * converted from integers to a string with the data type indicated in
     * the <code>:kril-wrapper</code> field of the reply message
     *
     * @see KrilWrapper
     */
    public Reply(String performative, int [] i)
    {
        this.performative   = performative;
        KrilWrapper wrapper = new KrilWrapper(i);
        keywordValues.put(CONTENT,      wrapper.getContent());
        keywordValues.put(KRIL_WRAPPER, wrapper.getKrilWrapper());
    }

    /**
     * Returns a Reply instance.
     *
     * @param performative  the reply message's performative 
     * @param i             the reply message's long value to be converted
     * from a long to a string with the data type indicated in the
     * <code>:kril-wrapper</code> field of the reply message
     * @exception InvalidLongInputException if the input long takes more
     * than 32 bits to be represented
     * @see KrilWrapper
     */
    public Reply(String performative, long l) throws InvalidLongInputException
    {
        this.performative   = performative;
        KrilWrapper wrapper = new KrilWrapper(l);
        keywordValues.put(CONTENT,      wrapper.getContent());
        keywordValues.put(KRIL_WRAPPER, wrapper.getKrilWrapper());
    }

    /**
     * Returns a Reply instance.
     *
     * @param performative  the reply message's performative 
     * @param l             the reply message's long array value to be
     * converted from longs to a string with the data type indicated in
     * the <code>:kril-wrapper</code> field of the reply message
     * @exception InvalidLongInputException if any the input longs takes more
     * than 32 bits to be represented
     * @see KrilWrapper
     */
    public Reply(String performative, long []l) throws InvalidLongInputException
    {
        this.performative   = performative;
        KrilWrapper wrapper = new KrilWrapper(l);
        keywordValues.put(CONTENT,      wrapper.getContent());
        keywordValues.put(KRIL_WRAPPER, wrapper.getKrilWrapper());
    }

    /**
     * Sets the value of the named keyword.
     * @param keyword the new keyword name
     * @param value the new keyword value
     * @return the previous string value for the keyword, null if the keyword
     * was not present
     */
    public String setKeywordValue(String keyword, String value)
    {
        return ((String)keywordValues.put(keyword, value));
    }

    //
    // Selector methods
    //
    /**
     * Selector returning the reply's performative
     *
     * @return the reply message's peformative
     */
    public String getPerformative() { return (performative); }

    /**
     * Selector returning the reply's <code>:content</code> field
     *
     * @return the reply message's <code>:content</code>
     */
    public String getContent() { return ((String)keywordValues.get(CONTENT)); }

    /**
     * Gets the value matching the keyword name
     * to the message.
     * @param keyword the new keyword name
     * @return the string value for the keyword, null if the keyword is not
     * present
     */
    public String getKeywordValue(String keyword)
    {
        return ((String)keywordValues.get(keyword));
    }

    //
    // The following are static methods to generate "Reply" objects with
    // a CANNED performative, such as "REPLY", "TELL", "SORRY" and "ERROR" 
    //
    // - different function signatures are provided and can be added to as
    //   is demanded by KRIL user programmers.
    //
    // - it is expected that the "SORRY" and "ERROR" performatives will only
    //   ever need to support a single string argument, while the "REPLY"
    //   and "TELL" functions will support int, long, String and arrays of
    //   the same types.
    //

    /**
     * Canned "REPLY" function
     */
    public static Reply reply(String content)
    {
        return (new Reply(REPLY, content));
    }

    /**
     * Canned "REPLY" function
     */
    public static Reply reply(String [] strings)
    {
        return (new Reply(REPLY, strings));
    }

    /**
     * Canned "REPLY" function
     */
    public static Reply reply(int i)
    {
        return (new Reply(REPLY, i));
    }

    /**
     * Canned "REPLY" function
     */
    public static Reply reply(int [] i)
    {
        return (new Reply(REPLY, i));
    }

    /**
     * Canned "REPLY" function
     * @exception InvalidLongInputException if the input long takes more
     * than 32 bits to be represented
     */
    public static Reply reply(long l) throws InvalidLongInputException
    {
        return (new Reply(REPLY, l));
    }

    /**
     * Canned "REPLY" function
     * @exception InvalidLongInputException if any the input longs takes more
     * than 32 bits to be represented
     */
    public static Reply reply(long [] l) throws InvalidLongInputException
    {
        return (new Reply(REPLY, l));
    }

    /**
     * Canned "TELL" function
     */
    public static Reply tell(String content)
    {
        return (new Reply(TELL, content));
    }

    /**
     * Canned "TELL" function
     */
    public static Reply tell(String [] strings)
    {
        return (new Reply(TELL, strings));
    }

    /**
     * Canned "TELL" function
     */
    public static Reply tell(int i)
    {
        return (new Reply(TELL, i));
    }

    /**
     * Canned "TELL" function
     */
    public static Reply tell(int [] i)
    {
        return (new Reply(TELL, i));
    }

    /**
     * Canned "TELL" function
     * @exception InvalidLongInputException if the input long takes more
     * than 32 bits to be represented
     */
    public static Reply tell(long l) throws InvalidLongInputException
    {
        return (new Reply(TELL, l));
    }

    /**
     * Canned "TELL" function
     * @exception InvalidLongInputException if any the input longs takes more
     * than 32 bits to be represented
     */
    public static Reply tell(long [] l) throws InvalidLongInputException
    {
        return (new Reply(TELL, l));
    }

    /**
     * Canned "ERROR" function
     */
    public static Reply error(String content)
    {
        return (new Reply(ERROR, content));
    }

    /**
     * Canned "SORRY" function
     */
    public static Reply sorry(String content)
    {
        return (new Reply(SORRY, content));
    }

    //
    // Generate the fields from the Reply object ... will usually
    // include a content field and a kril-wrapper field
    //
    protected void generateMessageText(StringBuffer textBuffer)
    {
        String name;
        String value;
        //
        // Generate the fields from the Reply object ... will usually
        // include a content field and a kril-wrapper field
        //
        for (Enumeration keys = keywordValues.keys(); keys.hasMoreElements();)
        {
            name  = (String)keys.nextElement();
            value = getKeywordValue(name);
            if (value != null && value.length() > 0)
                textBuffer.append(" "+name + " " + Message.quoteString(value));
        }
    }
}
