/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         TimingRecord.java
**
** Description:  Contains the TimingRecord class that is internal to the KRIL.
**
*/
package kqml;
import java.util.Date;

/**
 * A class internal to the KRIL.
 *
 * It it used to record and generate KQML message timing records for
 * transmission to the kqml timing agent.
 *
 * @author Peter Stevens
 * @version $Revision: 1.5 $
 *
 * @exception TimingRecordException if an invalid timing record type
 * is passed to a the constructor or methods.
 */

class TimingRecord
{
    protected static final int READ  = 1;
    protected static final int WRITE = 2;

    protected static final int TRY_AGAIN_PERIOD = 5000; // 5 seconds

    protected static final String TELL          = "tell";
    protected static final String TIMINGS_AGENT = "KQML-TiMinGs-AgeNT";


    //
    // Private Data
    //
    private static Date      timingsAgentDown      = null;
    private static boolean   timingsAgentHasBeenUp = false; 

    private String           desc             = null;
    private String           kqmlMsgId        = null;
    private long             miscElapsed      = 0;
    private Date             readStart        = null;
    private Date             readStop         = null;
    private int              type             = 0;
    private Date             writeStart       = null;
    private Date             writeStop        = null;


    TimingRecord(String kqmlMsgId, int type, Date start, Date stop)
                                            throws TimingRecordException
    {
        this.kqmlMsgId = kqmlMsgId;
        if ((this.type = type) == READ)
        {
            readStart  = start;
            readStop   = stop;
        }
        else if (this.type == WRITE)
        {
            writeStart = start;
            writeStop  = stop;
        }
        else
            throw new TimingRecordException (
                "TimingRecord: invalid timing record type " + type);
    }

    protected String getKqmlMsgId()  { return (kqmlMsgId);  }

    protected void completeTimingRecord(int type, Date start, Date stop)
                                throws TimingRecordException
    {

        if (type == READ)
        {
            desc        = "WAIT";
            readStart   = start;
            readStop    = stop;
            miscElapsed = readStart.getTime() - writeStop.getTime();
        }
        else if (type == WRITE)
        {
            desc        = "PROCESS";
            writeStart  = start;
            writeStop   = stop;
            miscElapsed = writeStart.getTime() - readStop.getTime();
        }
        else
            throw new TimingRecordException (
                "completeTiming: invalid timing record type " + type);
    }

    protected void sendTiming(KqmlAgent a)
    {
        String  content               = null;
        Message msg                   = null;
        long    readElapsed           = 0;
        long    writeElapsed          = 0;

        readElapsed  = readStop.getTime()  - readStart.getTime();
        writeElapsed = writeStop.getTime() - writeStart.getTime();

        content = kqmlMsgId + ":" + desc + "," + readElapsed + "," +
                                        writeElapsed + "," + miscElapsed;
        //
        // Now send this off to the Timings agent; WITHOUT any timing
        // recorded (no timing is recorded for OUTBOUND only messages)
        try
        {
            if (timingsAgentDown == null)
            {
                Log.log(Log.WRITE+Log.DEBUG, "sendTiming for id `" +
                            kqmlMsgId + "'\n    content is `" + content + "'");
                msg = new Message(TELL, content, TIMINGS_AGENT, a);
            }
        }
        catch (InvalidKQMLMsgException e)
        {
            Date now = new Date();

            if (timingsAgentDown == null)
            {
                if (timingsAgentHasBeenUp == true)
                    Log.log(Log.ERROR,"sendTiming: msg creation error on send");
                timingsAgentDown      = now;
                timingsAgentHasBeenUp = false;
            }
        }
        try
        {
            if (timingsAgentDown == null)
            {
                msg.send();
                timingsAgentDown      = null;
                timingsAgentHasBeenUp = true;
            }
        }
        catch (KqmlException e)
        {
            Date now = new Date();

            if (timingsAgentDown == null)
            {
                if (timingsAgentHasBeenUp == true)
                    Log.log(Log.ERROR, "sendTiming: IO Error on send");
                timingsAgentDown      = now;
                timingsAgentHasBeenUp = false;
            }
            else
                if ((now.getTime()-timingsAgentDown.getTime())
                                                            > TRY_AGAIN_PERIOD)
                {
                    timingsAgentDown = null;
                }
        }
    }

    //
    // "equals" and "hashCode" should return the same value for the
    // same identifier; these are present that that a TimingRecord
    // may be entered as an item in a Java HashTable.
    //
    public boolean equals (Object obj)
    {
        String otherMsgId;

        if (kqmlMsgId == null)
            return (false);
        else if ((otherMsgId = ((TimingRecord)obj).getKqmlMsgId()) == null)
            return (false);
        else
            return (kqmlMsgId.equals(otherMsgId));
    }

    public int hashCode ()
    {
        return (kqmlMsgId == null ?  super.hashCode() : kqmlMsgId.hashCode());
    }
}
