/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         UnsolicitedQueue.java
**
** Description:  Contains the UnsolicitedQueue class that is internal to
**               the KRIL.
**
*/
package kqml;
import java.util.Vector;

/**
 * A class internal to the KRIL.
 *
 * This class is used to implement a general Unsolicited KQML Message queue.
 * Currently a general unsolicited solicited message queue is used to feed
 * the message handler threads in order to pass a message to the first
 * idle handler thread when all message handler threads are active at
 * the time an unsolicited message is received.
 *
 * @author Peter Stevens
 * @version $Revision: 1.5 $
 * @see KqmlAgent
 * @see KqmlDispatch
 */
class UnsolicitedQueue
{
    private static int   lastThreadNo = 0;
    private Vector       queue        = new Vector();

    public void putMsg(Message msg, KqmlDispatch[] msgThreads)
    {
        int     j            = lastThreadNo++;
        boolean threadIsBusy = true;

        // Put a message on the unsolicited message queue
        synchronized (queue)
        {
            queue.addElement(msg);
        }

        // Resume a "non-busy" thread
        for (int i = 0; i < msgThreads.length && threadIsBusy; ++i, ++j)
        {
            if (j >= msgThreads.length)
                j = 0;

            threadIsBusy = msgThreads[j].isBusy();
            if (!threadIsBusy)
            {
                lastThreadNo = j + 1;
                msgThreads[j].resume();
                Log.log(Log.THREAD, "putMsg: unsolicited msg from `" +
			msg.getSender() +
			"'\n     will be handled by thread `" +
			msgThreads[j].getThread().getName() + "'");
            }
        }

        if (threadIsBusy)
        {
            Log.log(Log.THREAD, "putMsg: unsolicited msg from " +
		    msg.getReceiver() +
		    "\n     received with all threads busy");
        }
    }

    public Message getMsg()
    {
        Message rval = null;

        synchronized (queue)
        {
            if (!queue.isEmpty())
            {
                rval = (Message)queue.firstElement();
                queue.removeElement(rval);
            }
        }

        if (rval == null)
        {
            Thread.currentThread().suspend();

            synchronized (queue)
            {
                if (!queue.isEmpty())
                {
                    rval = (Message)queue.firstElement();
                    queue.removeElement(rval);
                }
            }
        }

        return (rval);
    }
}
