/* File: SendKqml.java
 * Contains: Main program for testing Kqml message sending
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Thu Apr 16 19:10:44 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.test;

import java.util.*;
import java.io.*;
import oplan.util.*;
import oplan.util.lisp.*;
import oplan.util.match.*;
import oplan.acp3.*;
import kqml.*;


/**
 * A class to test Kqml message sending.<p>
 *
 * The first command-line argument, if given, indicated what SendKqml
 * should call itself as a Kqml agent.
 */

class SendKqml implements KqmlSymbols {

    private static String ourname = "SendKqml";
    private static String serverName = "ACP3";
    private static KqmlAgent kqmlAgent = null;
    private static LispReader lispin = new LispReader(System.in);

    private static int delay = 0;

    private static String[] performatives_to_handle = {
	"reply", "error", "sorry", "evaluate", "tell"
    };

    /* Symbols */

    static final Symbol S_SENDLOG      = Symbol.intern("sendlog");
    static final Symbol S_DELAY        = Symbol.intern("delay");
    static final Symbol S_HELP         = Symbol.intern("help");
    static final Symbol S_BYE          = Symbol.intern("bye");

    /* Help text */

    static final String[] helpText = {
	"",
	"Sending to " + serverName,
	"",
	"If you type a list, it should have the form",
	"  (<kqml-performative> ... :content <the-content> ...)",
	"",
	"Such a list will be converted to a Kqml Message and sent.",
	"Commands have the form: <name> <argument>*, and the following",
	"commands are available:",
	"",
	"   ping <agent-name>         ping the specified agent",
	"   sendlog <log-file-name>   send the contents of the file to ACP3",
	"   delay <seconds>           set time between messages for sendlog",
        "   help                      print this text",
	"   bye                       exit",
	""
    };

    /* Main */

    public static void main(String[] argv) {

	if (argv.length == 1) {
	    ourname = argv[0];
	}
	System.out.println("Calling ourself \"" + ourname + "\"");

	// Create the Kqml agent.
        try { kqmlAgent = new KqmlAgent(ourname); }
        catch (KqmlAgentRegistrationFailure e) { fatalException(e); }
        catch (KqmlAgentInitFailure e)         { fatalException(e); }

	// Register handlers for messages we think we might receive
	{
	    for (int i = 0; i < performatives_to_handle.length; i++) {
		String perf = performatives_to_handle[i];
		kqmlAgent.registerHandler
		    (perf, new MessageNoter(perf, kqmlAgent));
	    }
	}

	// If we don't lower our priority, replies aren't printed
	// while we're at the "Message:" prompt.  /\/ No, that helps
	// only some of the time, and the behaviour is different on
	// Solaris and FreeBSD.

	// Thread main_thread = Thread.currentThread();
	// main_thread.setPriority(main_thread.getPriority() - 1);

	// Loop asking the user for a message and sending it

	Util.printLines(helpText);

    askloop:
	for (;;) {
	    try {
		Object input = askLisp("\nMessage:");
		if (input == S_BYE)
		    break askloop;
		else if (input instanceof Cons)
		    sendMessageFromList((Cons)input);
		else if (input == S_PING)
		    sendPing(lispin.readObject().toString());
		else if (input == S_SENDLOG)
		    sendLogFile(lispin.readObject().toString());
		else if (input == S_DELAY)
		    delay = ((Number)lispin.readObject()).intValue();
		else if (input == S_HELP)
		    Util.printLines(helpText);
		else
		    Util.printLines(new String[] {
			"I don't understand " + input,
			"Type \"help\" for help"
		    });
	    }
	    catch (KqmlException e) {
		fatalException(e);
	    }
	    catch (RuntimeException e) {  // s.b. LispReadException
		Debug.noteException(e);
	    }
	}

	// And we're done.
	kqmlAgent.shutdown();
	System.out.println("\nBye!\n");

    }

    private static void fatalException(Exception e) {
        System.out.println(e.getMessage());
	if (kqmlAgent != null)
	    kqmlAgent.shutdown();
	System.exit(1);
    }

    static Object askLisp(String prompt) {
	System.out.print(prompt + " ");
	System.out.flush();
	// It's annoying that there's no lispin.available().
	try {
	    while (System.in.available() == 0) {
		try { Thread.sleep(1000); }
		catch (InterruptedException e) {}
	    }
	    return lispin.readObject();
	}
	catch (IOException e) {
	    Debug.noteException(e);
	    return Lisp.NIL;
	}
    }

    /* Message sending (from lists)
     *
     * A message might be:
     *
     * (EVALUATE :SENDER "pcm-1" :REPLY-WITH :INIT-PROBLEM :CONTENT
     *   (:INIT-PROBLEM :TASK "AS-A" :PLAN "plan-advice1"))
     *
     */

    private static void sendMessageFromList(List message) 
	 throws KqmlException
    {
	sendMessageFromList(message, true);
    }

    private static void quietlySendMessageFromList(List message) 
	 throws KqmlException
    {
	sendMessageFromList(message, false);
    }

    private static void sendMessageFromList(List message, boolean trace)
	 throws KqmlException
    {
        if (trace) Debug.noteln("Have:", Lisp.printToString(message));

	String performative = message.car().toString();
	List plist = message.cdr();
	String content = Lisp.printToString(plist.get(K_CONTENT));

	if (trace) Debug.noteln("Performative:", performative);
	if (trace) Debug.noteln("Content:", content);

	Message m = new Message(performative,
				content,
				serverName, 	// to
				kqmlAgent);	// from

	// Fill in message properties, skipping :content.
	// If there's a :sender property in the message (as there is
	// in messages taken from log files), it will overwrite the
	// sender value that was put in when the Message was constructed.
	// So on the receiving end the Message will appear to come from
	// the original sender rather than from this program.  /\/

	for (List p = plist; !p.isNull(); p = p.cdr().cdr()) {
	    Debug.assert(p.car() instanceof Symbol);
	    if (p.car() != K_CONTENT) {
		String prop = p.car().toString();
		String value = p.cdr().car().toString();
		if (trace) Debug.noteln("Property " + prop + " = " + value);
		m.setKeywordValue(prop, value);
	    }
	}

	// Send the message
	m.send();

    }


    /* Ping a specified agent */

    /* If you want to send a ping by hand, the list form of the message
     * looks like this:
     *
     *    (evaluate :reply-with :ping :content (:ping))
     *
     * Here we use sendMsgAndAwaitReply.
     */

    static void sendPing(String who) throws KqmlException {
	try {
	    Message reply
		= kqmlAgent.sendMsgAndAwaitReply("evaluate", "(:ping)", who);
	    Debug.noteln("Reply:", reply);
	}
	catch (Exception e) {
	    Debug.noteException(e);
	}
    }


    /* Send the contents of a log file */

    static void sendLogFile(String filename) throws KqmlException {
	KqmlLogReader lin;
	Object message;

	try {
	    lin = KqmlLogReader.newInstance(filename);
	    while ((message = lin.nextMessage()) != Lisp.EOF) {
		quietlySendMessageFromList((List)message);
		try { Thread.sleep(1000 * delay); }
		catch (InterruptedException e) {}
	    }
	    System.out.println("\nContents of " + filename + " sent.");
	}
	catch (FileNotFoundException e) {
	    Debug.noteException(e);
	}

    }


}


/*
 * A simple handler class for messages we receive.
 */

class MessageNoter implements Handler {

    String performative;
    KqmlAgent kqmlAgent;

    MessageNoter(String performative, KqmlAgent kqmlAgent) {
	this.performative = performative;
	this.kqmlAgent = kqmlAgent;
    }

    public Reply handler(Message msg) {
	Debug.noteln("Received: " + msg.getPerformative() +
		     " from " + msg.getSender() +
		     ", content = " + msg.getContents());
	// Debug.noteln("Whole message: " + msg);
	if (msg.getContents().equals("die")) {
	    kqmlAgent.shutdown();
	    System.exit(0);
	}
	return makeReply(msg);
    }

    static protected Reply makeReply(Message msg) {
	// Is a reply called for?
	// The :reply-with field is automatically send back as :in-reply-to.
	// It's not clear what we'll ever get :reply-with t or nil.
	String reply_with = msg.getReplyWith();
	// Debug.noteln("Making reply to", msg);
	if (reply_with == null)
	    return null;
	else if (reply_with.equalsIgnoreCase("nil"))
	    return null;
	else if (reply_with.equalsIgnoreCase("t"))
	    return Reply.reply(":ok");
	else
	    return Reply.reply("(" + reply_with + " :ok)");
    }

}
